%{ /* Emacs, use -*- mode:text -*- */
/*
 * Lexically analyzes ausearch output.
 * John D. Ramsdell--The MITRE Corporation
 *
 * Copyright (C) 2006 The MITRE Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * The main type of token recognized is an item.  An item is made up
 * of non-white space and equal sign characters.  The analyzer treats
 * material within parentheses, double quotes, and single quotes as
 * part of an item.
 *
 * The lexical analyzer ignores time annotations, and treats records
 * of type=AVC specially, since they do not obey normal syntactic
 * rules.
*/

#include <stdio.h>
#include <unistd.h>
#include "scanner.h"

/* For error reporting. */
static int lineno = 1;

%}

S			[ \t]
ITEM			[^\n \t()"'=]|{STR}|{PAR}
STR			\"([^\"\\\n])*\"|\'([^\'\\\n])*\'
PAR			\(([^\(\)\\\n])*\)

%%

{S}+	   		{ }

\n			{ lineno++; return NL; }

=			{ return EQ; }

^----$			{ return SEP; }

^time->.*$              { }

^type=AVC.*$		{ return AVC; }

{ITEM}+			{ return ITEM; }

.			{ yyerror("Bad input token"); exit(1); }

%%

const char *
getyytext(void)
{
  return yytext;
}

static const char *filename = "-";

int
yyerror(const char *msg)
{
  return fprintf(stderr, "%s:%d: %s at token %s\n",
                 filename, lineno, msg, yytext);
}

void
setfile(const char *arg)
{
  lineno = 1;
  if (arg)
    filename = arg;
  else
    filename = "-";
}
