#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

static char buf[] = "Hello World!\n";

static int
say(int s)
{
  size_t n = sizeof(buf);

  size_t i = n;
  char *msg = buf;
  for (;;) {
    ssize_t j = write(s, msg, i);
    if (j < 0) {
      perror("Write");
      return 1;
    }
    if (j == 0) {
      fprintf(stderr, "Write blocked\n");
      return 1;
    }
    i -= j;
    if (i == 0)
      break;
    msg += j;
  }

  msg = buf;
  for (;;) {
    ssize_t k = read(s, msg, n);
    if (k < 0) {
      perror("Read");
      return 1;
    }
    n -= k;
    msg += k;
    if (n == 0)
      break;
  }

  *msg = 0;
  printf("%s", buf);
  return 0;
}

static int
conn(const char *host, int port)
{
  struct sockaddr_in sain[1];
  struct hostent *hostent = gethostbyname(host);

  if (!hostent) {
    fprintf(stderr, "Cannot look up %s\n", host);
    return 1;
  }

  memset(sain, 0, sizeof(*sain));
  sain->sin_family = hostent->h_addrtype;
  memcpy(&(sain->sin_addr), hostent->h_addr, hostent->h_length);
  sain->sin_port = htons((u_short)port);

  int s = socket(PF_INET, SOCK_STREAM, 0);
  if (s == -1) {
    perror("Socket");
    return 1;
  }

  if (!connect(s, (struct sockaddr *)sain, sizeof(*sain)) < 0) {
    perror("Connect");
    return 1;
  }

  return say(s);
}

int
main(int argc, char **argv)
{
  if (argc == 3)
    return conn(argv[1], atoi(argv[2]));
  fprintf(stderr, "Usage: %s host port\n", argv[0]);
  return 1;
}
