#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>

static char buf[] = "Hello World!\n";

static int
say(int s)
{
  size_t n = sizeof(buf);

  size_t i = n;
  char *msg = buf;
  for (;;) {
    ssize_t j = write(s, msg, i);
    if (j < 0) {
      perror("Write");
      return 1;
    }
    if (j == 0) {
      fprintf(stderr, "Write blocked\n");
      return 1;
    }
    i -= j;
    if (i == 0)
      break;
    msg += j;
  }

  msg = buf;
  for (;;) {
    ssize_t k = read(s, msg, n);
    if (k < 0) {
      perror("Read");
      return 1;
    }
    n -= k;
    msg += k;
    if (n == 0)
      break;
  }

  *msg = 0;
  printf("%s", buf);
  return 0;
}

static int
echo(int s)
{
  size_t n = sizeof(buf);
  char b[sizeof(buf) + 1];

  char *msg = b;
  for (;;) {
    ssize_t k = read(s, msg, n);
    if (k < 0) {
      perror("Read");
      return 1;
    }
    n -= k;
    msg += k;
    if (n == 0)
      break;
  }
  *msg = 0;

  msg = b;
  size_t i = sizeof(buf);
  for (;;) {
    ssize_t j = write(s, msg, i);
    if (j < 0) {
      perror("Write");
      return 1;
    }
    if (j == 0) {
      fprintf(stderr, "Write blocked\n");
      return 1;
    }
    i -= j;
    if (i == 0)
      break;
    msg += j;
  }

  return 0;
}

static void *			/* Never dereference this result */
child(void *arg)
{
  int s = *((int *)arg);
  int rc = echo(s);
  return rc ? child : 0;
}

static int
conn(void)
{
  int s[2];
  if (socketpair(PF_UNIX, SOCK_STREAM, 0, s) == -1) {
    perror("Socketpair");
    return 1;
  }

  pthread_t thread;
  if (pthread_create(&thread, NULL, child, s + 1)) {
    fprintf(stderr, "Thread create error");
    return 1;
  }

  if (say(s[0]))
    return 1;

  void *rc = NULL;
  if (pthread_join(thread, &rc)) {
    fprintf(stderr, "Thread join error");
    return 1;
  }

  return rc != NULL;
}

int
main(int argc, char **argv)
{
  if (argc == 1)
    return conn();
  fprintf(stderr, "Usage: %s\n", argv[0]);
  return 1;
}
