from util import *
import domain_template
questions_dict = {}
if_dict = {}

interface_needs = []

class AppProperty:
    def __init__(self, name, brief_desc, active_flag = 1):
        self.name = name
        self.brief_desc = brief_desc
        self.active_flag = active_flag        

class Question (AppProperty):
    def __init__(self, name, desc, node_test = None, active_flag = 1):
        AppProperty.__init__(self, name, desc, active_flag)
        questions_dict[name] = self
        self.response = eval('domain_template.response_' + name)
        self.node_test = node_test
    def is_it_active (self, obj):
        toreturn = self.active_flag
        #now check to see if there is reason to change it
        if self.name in obj.__dict__.keys():
            there = getattr(obj, self.name)
            if there:
               toreturn = 1
        elif self.node_test != None:
            if self.node_test(obj):
                toreturn = 1
        return toreturn



################ node tests ###############################

def test_tcp (obj):
    return Exists(obj.get_extended_connects_at())

def test_exec (obj):
    return Exists(obj.get_extended_execs())

def test_read (obj):
     return Exists(obj.get_extended_reads())

def test_write (obj):
    return Exists(obj.get_extended_writes())

def test_var (obj):
    return Exists([x for x in obj.get_extended_reads() \
                   if x.find('/var') == 0])

def test_var_lib (obj):
    return Exists([x for x in obj.get_extended_reads() \
                   if x.find('/var/lib') == 0])

def test_etc (obj):
    return Exists([x for x in obj.get_extended_reads() \
                   if x.find('/etc') == 0])

    
############### Questions ################################
                   
Question('tcp','use tcp socket', test_tcp,  0)
Question('udp','use udp socket', None, 0)
Question('raw','use raw sockets', None, 0)
Question('exec_processes','execute processes', test_exec)
Question('read','read from files', test_read, 0)
Question('write','write to files', test_write, 0)
Question('var','access /var', test_var, 0)
Question('var_lib','access /var/lib', test_var_lib, 0)
Question('etc', 'access /etc', test_etc, 0)
Question('symlinks', 'access /var/lib/symlinks', None, 0)
Question('manage_logs', 'allow log management', None,  0)
Question('localization', 'access localization', None, 0)


################### interface calls #####################

class IF (AppProperty):
    """Prototypes"""
    reverse_allows = []
    allows = []
    macros = []
    required_types = []
    def __init__(self, name, desc, active_flag = 1):
        AppProperty.__init__(self, name, desc, active_flag)
        if_dict[name] = self

        
class  DomainTransIF (IF):
    allows = ['t:fd use']
    macros = [['domain_auto_trans', ['$1', 't', 'exec_t']]]
    reverse_allows = [['t', 'fd use'], ['data_t', 'fifo_file rw_file_perms'],
                      ['t', 'process sigchld']]                                             
    params = ['domain']
    required_types = ['exec_t', 't']
_dt = DomainTransIF('domain_trans', 'Can be executed from own domain')
        
class  ExecIF (IF):
    macros = [['can_exec', ['$1', 'exec_t']]]
    params = []
    required_types = ['exec_t']    
_e = ExecIF('exec', 'Others can execute')
        
class SignalIF(IF):
    allows = ['t:process signal']
    params = []
    required_types = ['t']    
_sig = SignalIF('signal', 'Others can signal', 0)
        
class  SignullIF(IF):
    allows = ['t:process signull']
    params = []
    required_types = ['t']    
_signull = SignullIF('signull', 'Others can send a null signal', 0)

    
class  SigchldIF (IF):
    allows = ['t:process sigchld']
    params = []
    required_types = ['t']    
_sigchld = SigchldIF('sigchld','Others can send a sigchld signal', 0)
        
class UseFDIF (IF):
    allows = ['t:fd use']
    params = []
    required_types = ['t']    
_usefd = UseFDIF('use_fd', 'Others can use file descriptors', 0)
        
class  ReadConfigIF(IF):
    allows = ['config_t:dir r_dir_perms', \
    'config_t:file r_file_perms']
    params = []
    required_types = ['config_t']    
_rconfig = ReadConfigIF('read_config', 'Others can read configuration files', 0)
        
class  ReadLogIF(IF):
    allows = ['log_t:dir r_dir_perms', 'log_t:file r_file_perms', \
    'log_t:lnk_file { getattr read }']
    params = []
    required_types = ['log_t']    
_rlog = ReadLogIF('read_log', 'Others can read log files', 0)
        
class  AppendIF(IF):
    allows = ['data_t:dir ra_dir_perms', 'file { create append getattr }']
    params = []
    required_types = ['data_t']    
_append = AppendIF('append', 'Others can append to files', 0)
        
class  CanRWConfig(IF):
    allows = ['config_t:dir ra_dir_perms', \
    'config_t:file { getattr read write create }']
    params = []
    required_types = ['config_t']    
_rwconfig = CanRWConfig('can_r_w_config', \
                        'Others can read and write configuration files', 0)
        
class  SearchIF(IF):
    allows = ['data_t:dir search_dir_perms']
    params = []
    required_types = ['data_t']    
_search = SearchIF('search', 'Others can search data directories', 0)
        
class  ListDataIF (IF):
    allows = ['data_t:dir r_dir_perms']
    params = []
    required_types = ['data_t']    
_list = ListDataIF('list_data', 'Let others list contents of data directories',0)
        
class  ReadDataSymlinks(IF):
    allows = ['data_t:dir search', 'data_t:lnk_file read']
    params = []
    required_types = ['data_t']    
_rsym = ReadDataSymlinks('read_data_symlinks', \
                         'Allow read access to data symbolic links', 0)

class  ManageLogIF(IF):
    allows = ['log_t:dir rw_dir_perms' , 'log_t:file create_file_perms' , \
              'log_t:lnk_file create_lnk_perms']
    params = []
    required_types = ['log_t']    
_managelog = ManageLogIF('manage_log', 'Allow others to read, write, delete logs', 0)
        
class  ReadIF (IF):
    allows = ['data_t:dir list_dir_perms', 'data_t:file r_file_perms', \
    'data_t:lnk_file { getattr read }']
    params = []
    required_types = ['data_t']    
_read = ReadIF('read', 'Allow others to read data files', 0)

