
import os
import stat
from util import actions, empty, Exists, contains_str, PRGM
import config

modules = {}

module_if_nodes = {}

search_requests = {}

class Module:
    def __init__(self, name):
        self.name = name
        modules[name] = self
        self.calls = []

class IF_Node:
    def __init__(self, category, module, call):
        #print "call is " + call
        self.category = category
        self.module = module
        self.call = call
        self.allows = []
        id = module + "_" + call
        self.id = id
        if id in module_if_nodes.keys():
            print "duplicate node at " + id
        module_if_nodes[id] = self
        if module not in modules.keys():
            modules[module] = Module(module)
        modules[module].calls.append(self)
    def __repr__(self):
        return "Node " + self.id

class  AllowInCall:
    def __init__(self, parent_call, str):
        """ str of the form
        allow $1 NetworkManager_t:udp_socket { read write };"""
        self.parent = parent_call
        fields = str.split()
        self.domain = fields[1]
        range_desc = str[7 + len(self.domain):]
        pos = range_desc.find(':')
        self.range = range_desc[: pos]
        rest = range_desc[pos + 1:]
        rest_list = rest.split()
        self.range_type = rest_list[0]
        self.perms = rest[len(self.range_type) + 1:-1]
    def __repr__(self):
        return "allow " + self.domain + " " + self.range + ":" + \
               self.range_type + " " + self.perms + ";"

        
class IF_Reader:
    def __init__(self, domain, filename, category):
        self.category = category
        self.domain = domain
        self.file = filename
        
    def read_it(self):
        fd = open(self.file, 'r')
        new = False
        in_require = 0
        for line in fd.readlines():
            l = line[:-1].strip()
            if l.find('interface') == 0:
                #have new interface call"
                pos1 = l.find("`")
                pos2 = l.find(',')
                title = l[pos1 + 1: pos2 - 1]
                #title = l.split(self.domain)[1][1:-2]
                obj = IF_Node(self.category, self.domain, title)
                new = True
            elif l.find('gen_require') != -1:
               in_require = 1
            elif in_require:
                if l.find("'") != -1:
                    in_require = 0
                else:
                    continue
            #elif l == "" or l.find('#') == 0:
                continue
            elif not in_require:
                if l.find("')") == 0:
                    new = False
                elif new and l.find('allow') == 0:
                    allow = AllowInCall(self, l)
                    #print allow
                    obj.allows.append(allow)
            else:
                print "something's wrong"


def visit_if_files():
    dir = '/usr/share/selinux/devel/include/'
    try:
        print os.listdir(dir)
    except:
        dir = config.prgmPath + '/modules/'
    paths = {}
    #print "Modules from " + dir
    try:
      subdirs = os.listdir(dir)
    except os.error, msg:
                print msg
                return 
    for d in subdirs:
        if stat.S_ISDIR(os.stat(dir + d)[0]) != 0:
            #print d
            paths[d] = dir + d + '/'
    for category in paths.keys():
        path = paths[category]
        #print "#####" + category + "######"
        for f in [f for f in os.listdir(path) if f[-3:] == '.if' and f[0] not in ['.', '#']]:
            IF_Reader(f[:-3], path + f, category).read_it()
            
class SearchRequest:
    def __init__(self, permissions,  object_type = '?', actor_type = '?', module = '?', exact_match = False):
        self.permissions = permissions
        self.exact = exact_match
        self.module = module
        self.actor = actor_type
        self.object = object_type
        search_requests[permissions] = self
    
def search (search_request):
    """Dispatch on the type of request"""
    permissions = search_request.permissions
    exact = search_request.exact
    module = search_request.module
    actor = search_request.actor
    object = search_request.object
    all = []
    if module == '?' or module not in modules.keys(): #match on any module
           search_set = modules.keys()
    else:
        search_set = [module]
    for m in search_set:
        #print m
        all = all + search_aux(m, permissions, object, actor, exact)
    return all


def search_aux(module, permissions, object, actor, exact_match):
    toreturn = []
    for  call in modules[module].calls:
        #print call.id
        allows = call.allows
        for a in allows:
            domain = a.domain #$1 is a wildcard
            range = a.range # might be a set e.g., {lib_t shlib_t}
             #print type + ' in ' + str([domain, range]) + '?'
            d_good =  match_it(domain, actor)
            r_good = match_it(range, object)
            if d_good and r_good:
                aliases = [permissions] + actions[permissions].friends
                if not exact_match and not empty([x for x in aliases if x in a.perms]) or \
                   exact_match and a.perms in [permissions] + actions[permissions].friends:
                    toreturn.append(call)
                    #print a
                    break
    return toreturn

def lst_from_str (types):
    x = types.replace('{', '')
    y = x.replace('}', '')
    z = y.strip()
    return z.split()

def match_it(types, key):
    """types is a string; target_lst is of the form  [r1_t, r2_t. ...]
    ri_t may be of the form x_t or x_$1_t"""
    target_lst = lst_from_str(types)
    if key == '?':
       return True
    if '$1' in target_lst: #$1 is a wildcard
        return True
    ok = False
    for r in target_lst:
        if contains_str(r,'_$'):
            segments = r.split('_')
            new = ''
            for s in segments:
                if s.find('$') != -1:
                    continue
                else:
                    new = new +  s + '_'
            r = new[:-1]
        if  r ==  key:
            ok = True
            break
    return ok

