from util import *
from operators import *
import pr
import patterns
from specs import SPECIFIERS
import os.path
import config
from domain_model import setup_normalized_transitions, actions
import presentations
import walk
from textentry import group_name_map, do_anomaly
from futil import reads_from, writes_to
import refpol
import extract_ifs
from domain_template import *
from psl_parsing import nodes
from dutil import questions_dict, if_dict, interface_needs

""" Callbacks from GUI - includes the policy generation code"""

types_declared = []

module_category_order = ['kernel', 'system', 'services', 'admin', 'apps']
ifs_to_use = {}
###################### Handlers for html or text generation ###############

class PrintRequest:
    def __repr__ (self):
        return "print request"

class HtmlPR (PrintRequest):
    def __init__(self):
        self.linefeed = '<br>'
        self.start= '<h3>'
        self.end = '</h3>'
        self.focus = '<font color = "104bcl">'
        self.end_focus = '</font>'
        self.intro = '<html> \n'
        self.ending = '</html>'

class TePR (PrintRequest):
    def __init__(self):
        self.linefeed =  ''
        self.start= ''
        self.end = ''
        self.focus = ''
        self.end_focus = ''
        self.intro = ''
        self.ending = ''

## prototypes       
html_request = HtmlPR()
text_request = TePR()

###################### Generator Classes ################################

class TeStrategy:
    
    def __init__(self, printreq):
        """ printreq dispatch on text vs. html"""
        self.name = printreq
        self.linefeed = printreq.linefeed
        self.start= printreq.start
        self.end = printreq.end
        self.focus = printreq.focus
        self.end_focus = printreq.end_focus
        self.intro = printreq.intro
        self.ending = printreq.ending

    def gen_te_page (self, out):
        """out is strm"""
        out.write(self.intro)
        out.write('\npolicy_module(' + get_current_program().name + ', 1.0.0) \n\n')
        self.do_types(out)
        self.do_allows(out)
        self.do_patterns(out)
        out.write(self.ending)

    def types_to_work_with (self):
        found = {}
        for r in resources_dict.values():
            orig = r.original_context
            now = r.get_sc()
            if now == None:
                print "security context not set for" + r
            #at this point temp types have been converted to program_name
            # following looks for situations where the change has occurred
            if orig != now and now.find(PRGM.name_for_type) != -1:
                if now in found.keys():
                    found[now].append(r)
                else:
                    found[now] = [r]
        return found
        
    def do_types (self, out):
        linefeed = self.linefeed
        start= self.start
        end = self.end
        focus = self.focus
        end_focus = self.end_focus
        out.write(start + "##################### Type Declarations ####################" \
                  + end + " \n\n")
        use = []
        found = self.types_to_work_with()
        debug_check(found.keys())
        for sc in found.keys():
            types = {}
            for support in found[sc]:
                ty = the_type(support).lower()
                if support.special_name != None:
                    ty = support.special_name
                if ty in types.keys():
                    types[ty].append(support)
                else:
                    types[ty] = [support]
            for t in types.keys():
                if t == 'spawn':
                    continue
                elif t in ['file', 'dir']:
                    if sc.find('exec_t') != -1:
                        attribute = 'file_type, exec_type'
                    else:
                        attribute = 'file_type, sysadmfile'
                elif t == 'process':
                    special = 0
                    for s in types[t]:
                         if s.entry_point == "1":
                            attribute = 'special'
                            special = 1
                            break
                    if special == 0:
                        attribute = 'domain'
                else:
                    attribute = t
                types_declared.append(context2type(sc))
                use.append(self.type_output(sc, attribute))
        # this removes some duplicates; but may be duplicates
        # due to multiple line entries
 
        use = seq2set(use)
        for item in use:
            out.write(linefeed + focus + item + end_focus + ' \n')

    def type_output (self, sc, attribute):
        if attribute == 'special':
            attribute ='domain'
        return 'type ' + context2type(str(sc)) + ',' \
                                      + attribute + '; \n\n'
    def do_allows(self, out):
        linefeed = self.linefeed
        focus = self.focus
        end_focus = self.end_focus
        pname = get_current_program().name
        out.write(self.start + "##################### Allow Statements ####################" + self.end + "\n\n")
        for nt in normalized_trans.values():
##            if nt.domain_context.find('polgen_temp') != -1 \
##               or nt.range_context.find('polgen_temp') != -1:
##                    # exists old keys - should not be here
##                    continue
##                  #see if there exists a pattern that reports this connection
            if nt.action not in ['spawn','activate', 'exec', 'clone']:
                if not self.is_covered(nt):
                    self.gen_allow_and_evidence (out, nt)
        self.do_properties(out)
        if len(ifs_to_use.keys()) > 0:
            out.write(linefeed + '##### connections to other modules  \n')
        for interface in ifs_to_use.keys():
            domains = []
            for item in ifs_to_use[interface]:
                text = str(item)
                words = text.split()
                d = words[1]
                if d not in domains:
                    domains.append(d)
            for d in domains:
                out.write(linefeed + focus + interface + \
                          '(' + d + ')' + end_focus + '\n')
            out.write(linefeed + '#---- because \n')
            for item in ifs_to_use[interface]:
                out.write(linefeed + '#-------' + str(item))

    def is_covered (self, nt):
        return nt.is_covered()

    def do_properties(self, out):
        return "done"
            
    def do_patterns (self, out):
        linefeed = self.linefeed
        start = self.start
        end = self.end
        focus = self.focus
        end_focus = self.end_focus
        out.write(start + "##################### Pattern Instances ####################" + end + "\n\n")
        instances = [pa for pa in pr.pattern_assertions_final.values() if pa.accepted == 1]
        for patstr in patterns.PATTERN_TYPES:
            current = [pa for pa in instances if pa.pattern.name == patstr]
            if len(current) != 0:
                for v in current:
                    realization = pr.all_pattern_realizations[v.evidence[0]]
                    if self.pattern_ok(realization):
                         out.write(linefeed + focus + realization.spec() \
                               + end_focus + '\n')                        
                         for e in v.evidence: #set of realizations:
                                out.write(linefeed + '#--------------------' + \
                                          str([x.name for x in pr.all_pattern_realizations[e].constellation[:-1]]) + '\n')
                         out.write(' \n')

    def gen_allow_and_evidence (self, out, nt):
        linefeed = self.linefeed
        focus = self.focus
        end_focus = self.end_focus
        out.write(focus + linefeed + str(nt) + end_focus + ' \n')
        out.write(linefeed + '#--------------because \n')
        for e in nt.evidence_for:
            out.write(linefeed + '#-----------------' + str(e.name) +
                      '\n')

def debug_check(found):
    types = [context2type(x) for x in found]
    all = [x for x in resources_dict.values()]
    for x in all:
        l = context2type(x.security_context)
        if contains_str(l, get_current_program().name):
            if l not in types:
                print 'Problem: ' + x.name + ' with ' + l \
                      + " not covered in .te"
    
class TeMonolithic (TeStrategy):

    def pattern_ok (self, realization):
        if realization.contains_external():
                get_current_program().external_connections.append(realization.spec())
        return True



class TeReference (TeStrategy):
    
    def pattern_ok (self, realization):
        toreturn = True
        if realization.contains_external():
                get_current_program().external_connections.append(realization.spec())
                toreturn = False
        return toreturn


    def type_output(self, sc, attribute):
        """dispatch on attribute to determine file interface call to use"""
        ty = context2type(str(sc))
        # avoid duplicates
        if contains_str(ty, 'exec_t'):
           return ""
        add_exec = ""
        add_domain_entry_file = ""
        if attribute == 'special':
            interface = 'domain_type'
            exec_file = ty[:-2] + '_exec_t'
            # this should get picked up in other type declarations
            #add_exec = 'type ' + exec_file + ';\n' + self.linefeed
            add_exec = "type " + exec_file + '; \n' + self.linefeed
            add_domain_entry_file = 'domain_entry_file(' + ty + ', ' \
                                    + exec_file + ') \n' + self.linefeed
        elif ty == get_current_program().name + '_exec_t':
            return 'type ' + ty + '; \n'
        else:
            interface = 'files_type'
        return 'type ' + ty + '; \n' + self.linefeed + \
               add_exec + add_domain_entry_file + \
               interface + '(' + ty + ') \n\n'
        
    def is_covered (self, nt):
        """more complex - need to ask if there is a 'local' pattern that covers the
        information flow"""
        coveredp = 0
        if nt.domain_context == 'tbd' or nt.range_context == 'tbd':
            return 0
        #see if there exists a pattern assertion with a realization
        #using this transition
        evidence = nt.evidence_for
        for e in evidence:
            if len(e.marked) > 0:
                if not e.contains_external():
                    coveredp = 1
                    break
        return coveredp
                            
    def do_properties(self, out):
        #print "In do_properties"
        linefeed = self.linefeed
        focus = self.focus
        end_focus = self.end_focus
        special_nodes = [x for x in nodes.values() if x.entry_point == '1']
        if len(special_nodes) > 0:
            out.write(linefeed + linefeed + '\n##### The following are generated from data in the psl file \n')
            for n in special_nodes:
                out.write(linefeed + '### for ' + n.name + ' \n')
                #out.write(linefeed + focus + str(response_gen_basic(PRGM.name, n.name)) + end_focus)
                for q in questions_dict.keys():
                    if getattr(n,q): 
                        lines = questions_dict[q].response(PRGM.name, n.name)
##                        print str(n) + "-->" + str(q)
##                        print lines
                        for ln in lines:
                            out.write(linefeed + focus + ln + end_focus)
                out.write(linefeed + ' \n')
        return "done"

        
    def gen_allow_and_evidence (self, out, nt):
        linefeed = self.linefeed
        focus = self.focus
        end_focus = self.end_focus
        ok = True
        actor = context2type(nt.domain_context)
        object = context2type(nt.range_context)
        action = actions[nt.action]
        if action.direction == 'in':
            switch = object
            object = actor
            actor = switch
        actor_local =  contains_str(actor, PRGM.name_for_type)
        object_local = contains_str(object, PRGM.name_for_type)
        if not actor_local:
            ok = False
        elif  object != 'self':
            if  not object_local:
                ok = False
        #print "ok is " + str(ok)
        if ok:  # the allow only involves local types
           TeStrategy.gen_allow_and_evidence(self,out,nt)
        else: # external types involved
            calls = []
            dispatch_nbr = 0
            if object != 'self':
                if actor_local: # local domain, range must be non-local
                    dispatch_nbr = 1
                    calls = extract_ifs.search(extract_ifs.SearchRequest(nt.action, object, '?', '?', False))
                elif not object_local:  # both domain and range are from another module
                    # might not want to do this
                    dispatch_nbr = 2
                    calls = extract_ifs.search(extract_ifs.SearchRequest(nt.action, object, actor, '?', False))
                else:  #local range, but non-local domain
                    dispatch_nbr = 3
                    calls = extract_ifs.search(extract_ifs.SearchRequest(nt.action,'?', actor, '?',  False))
            else: # object != 'self':
                if not actor_local: # this is only case
                    dispatch_nbr = 4
                    calls = extract_ifs.search(extract_ifs.SearchRequest(nt.action, object, actor, None, False))
                
            if len(calls) == 0:
                print "$$$ Cannot find an interface call for " + str(nt)
                print "$$$ Have refpolicy modules been loaded"
                add_to_interface_needs(nt)
            elif len(calls) == 1:
                #print "found one call and id is " + calls[0].call
                id = calls[0].call
                if id in ifs_to_use.keys():
                    ifs_to_use[id].append(nt)
                else:
                    ifs_to_use[id] = [nt]
            else:
                bcall = best_choice(nt, calls)
                #print "best choice is " + bcall.call
                if bcall != None:
                    id = bcall.call
                    #print "add to collection " + id
                    if id in ifs_to_use.keys():
                        ifs_to_use[id].append(nt)
                    else:
                        ifs_to_use[id] = [nt]
        #print "ifs_to_use = " + str(ifs_to_use.keys())

def add_to_interface_needs (nt):
    if nt not in interface_needs:
        interface_needs.append(nt)
    
class TeTemp (TeReference):

    # In TeStrategy; need times for all converted types. At this point temp
    # types have been converted to program_name and the code is
    # following looks for situations where the change has occurred
    # if orig != now and now.find(PRGM.name_for_type) != -1:
    # Here we may need to catch situations where a pattern-based type has been set
    # e.g., polgen_temp_data_t -> ecommerce_pipeline_t
    def types_to_work_with (self):
        found = {}
        for r in resources_dict.values():
            orig = r.original_context
            now = r.get_sc()
            if now == None:
                    print "security context not set for" + r
            if contains_str(orig, 'polgen_temp'):
                if now in found.keys():
                    found[now].append(r)
                else:
                    found[now] = [r]
        return found

    def do_allows(self, out):
        linefeed = self.linefeed
        focus = self.focus
        out.write(self.start + "##################### Allow Statements ####################" + self.end + "\n\n")
        for nt in normalized_trans.values():
            if nt.action not in ['spawn','activate', 'exec', 'clone']:
##                if not self.is_covered(nt):
                    print nt
                    self.gen_allow_and_evidence (out, nt)
    
    def gen_te_page (self, out):
        """out is strm"""
        out.write(self.intro)
        out.write('\npolicy_module(' + get_current_program().name + ',1.0.0) \n\n')
        self.do_types(out)
        self.do_allows(out)
        self.do_patterns(out)
        out.write(self.ending)
        
temonotext = TeMonolithic(text_request)

temonohtml = TeMonolithic(html_request)

tereftext = TeReference(text_request)

terefhtml = TeReference(html_request)



class GenerationStrategy:

    def __init__ (self):
        self.files = {}

    def do_if (self, strm, format):
        genif = refpol.GenIF()
        genif.do_it_all(strm, refpol.M4Writer(PRGM.name), format, types_declared)

    def do_reports (self, format_request, ext_on_ext = ""):
        """ text for .te (experimental), .te, and .if"""
        if format_request == html_request:
            regobj = temonohtml
            refobj = terefhtml
        else:
            regobj = temonotext
            refobj = tereftext
        out_te = self.open_file('experimental_te' + ext_on_ext )
        regobj.gen_te_page(out_te)
        out_ref_te = self.open_file('te' + ext_on_ext)
        refobj.gen_te_page(out_ref_te)
        out_if = self.open_file('if' + ext_on_ext)
        self.do_if(out_if, format_request)
        
    def do_text_pages (self):
        """ text for .te (experimental), .te, and .if"""
        self.do_reports(text_request)

    def do_web_pages (self):
        """ web pages for .te (experimental), .te, and .if"""
        self.do_reports(html_request, ".html")
        
    def open_file (self, extension):
        prgm = get_current_program()
        pname = prgm.name
        fileid  = pname + "." + extension
        fileStr = config.result_file(fileid)
        out = open(fileStr, "w")
        print "Writing to " + fileStr
        self.files[extension] = out
        return out

      
    def do_allows_and_macros (self, strm, test_for_coveragep):
        return "done"
    
class PolgenTemp (GenerationStrategy):
    def open_file (self, extension):
            prgm = get_current_program()
            pname = prgm.name
            fileid  = 'temp_' + pname + extension
            fileStr = config.result_file(fileid)
            out = open(fileStr, "w")
            print "Writing to " + fileStr
            self.files[extension] = out
            return out
    
    def do_reports (self, format_request):
        """ text for .te (experimental), .te, and .if"""
        regobj = TeTemp(text_request)
        #refobj = tereftext
        out_te = self.open_file('.te' )
        regobj.gen_te_page(out_te)
##        out_ref_te = self.open_file('te' + ext_on_ext)
##        refobj.gen_te_page(out_ref_te)
        out_if = self.open_file('.if')
        self.do_if(out_if, format_request)

    def do_if (self, out, format_request):
        pol = refpol.GenIF()
        pol.ty_declared = types_declared
        pol.do_interface(out, if_dict['domain_trans'],
                         refpol.M4Writer('polgen_temp', 'm4_domain_trans'),
                         format_request)

        
class NoMacros (GenerationStrategy):
    
    def do_allows_and_macros(self, out, test_for_coveragep = 0): #current with test_for_coverage = 1
        seen = []
        out.write("\n\n##################### Allow Statements ##################### \n\n\n")
        populate_marks_on_transitions()
        for nt in normalized_trans.values():
            if nt.domain_context.find('polgen_temp') != -1 or nt.range_context.find('polgen_temp') != -1:
                # exists old keys - should not be here
                continue
            #current - only do next if not nt.is_covered()
            #see if there exists a pattern that reports this connection
            if not test_for_coveragep or not nt.is_covered():
                if nt.action != 'spawn':
                    if nt.action not in ['activate', 'exec']:
                        out.write(str(nt) + '\n')
                        out.write('#         because \n')
                        for e in nt.evidence_for:
                            out.write('#            ' + str(e) + '\n')



class WithMacros (GenerationStrategy):
    #obsolete
    def do_allows_and_macros(self, strm, test_for_coveragep): #this is current stategy
        the_no_macro_strategy.do_allows_and_macros(strm, 1)
        strm.write("##################### Pattern Instances #################### \n\n")
        instances = [pa for pa in pr.pattern_assertions_final.values() if pa.accepted == 1]
        for patstr in patterns.PATTERN_TYPES:
            current = [pa for pa in instances if pa.pattern.name == patstr]            #print len(current)
            if len(current) != 0:
                for v in current:
                    realization = pr.all_pattern_realizations[v.evidence[0]]
                    # print realization.contains_external()
                    
                    strm.write(realization.spec() + '\n')
                    #strm.write(v.print_form)
                    for e in v.evidence: #set of realizations:
                        strm.write('#          because of  ' + \
                                   str(pr.all_pattern_realizations[e].constellation[:-1]) + '\n')
                    strm.write(' \n')


class WithMacrosForSurvivorsOnly (GenerationStrategy):
    def do_allows_and_macros(self):
        return "done"

class TypingStrategy:
    def set_sc_for_objCB(self, f, value):
        """called via gui - value determined by user interaction"""
        TY_From_Contributors().go(f, value)


######################## Prototypes #######################################

TS = TypingStrategy()

gen_strategy = WithMacros()

the_no_macro_strategy = NoMacros()



##################### Callbacks #####################################

def adjust_pattern_list_finalCB ():
    """we look at active status of pattern realizations
    and adjust the type contributors
    At this point realizations should have their active property
    set (either by user or by default)"""
    for p in pr.all_pattern_realizations.values():
        patname = the_type(p)[:-1]
        contributor = patname.lower()
        patobj = eval('patterns.' + patname)
        participants = patobj.participants
        #first make sure any type modifications are removed if p is not active
        can_modify_type = patobj.gen_resource_types
        if can_modify_type != 'no':
            for part in participants:
                obj = p.__dict__[part] # the resource object
                if the_type(obj) == 'File' or \
                   (the_type(obj) == 'Process' and can_modify_type == part):
                    if p.active == 0 and contributor in obj.type_contributors:
                        obj.type_contributors.remove(contributor)
                        TY_Replace().go(obj, context2type(obj.original_context))
                    elif p.active == 1 and contributor not in obj.type_contributors:
                        obj.type_contributors.append(contributor)
      
                    
    




def set_integrity_groupCB(process_name, group):
    resources_dict[process_name + ':Process'].group = group

def set_active_valueCB (item, value):
    v = 1
    if value == 'OFF':
        v = 0 # delete pattern instance
    TY_Adjust_Active().go(pr.pattern_assertions[item], v, "User choice")




def set_file_contextsCB (group, val):
       if val != 'ignore':
        processes = [p for p in resources_dict.values() if the_type(p) == 'Process' and \
                     p.group == group]
        files = []
        if val in ['ALL files read', 'ALL files read or modified']:
            for p in processes:
                there = [t.domain for t in p.in_edges  \
                         if the_type(t.domain) == 'File']
                files = files + there
        elif val in ['ALL files modified', 'ALL files read or modified']:
            for p in processes:
                there = [t.range for t in p.out_edges \
                         if the_type(t.range) == 'File']
                files = files + there
        for f in seq2set(files):
            post(f, 'type_contributors', group)
            f.type_contributors.append(group)

def set_scCB (name, value, return_to_original = 0):
    f = resources_dict[name + ':File']
    TS.set_sc_for_objCB(f, value)

def set_dir_scCB (name, value):
    obj = resources_dict[name + ':File']
    post(obj, 'sc_extent', value)
    obj.sc_extent = value




##################### Policy Generators #########################
    



### typing rules
## original context could be set by environment or equal the polgen default
## if the polgen default then it is possible that the type could be changed thru participation in a pattern
##      the new type would be <prgm-name>_<pattern-type>_t
## or thru belonging to a group
##      the new type would be <prgm-name>-<group-name>_t
## environment types are not reset by polgen -
##    that is, we only want to change types for those files that are "owned" by the application
##    if we changed a type from 1_t to 2_t, then every other application that used to communicate
##    with 1_t would now also need to communicate with 2_t - should we allow the policy writer to do this?
### let's say we allow the policy writer to change what he/she wants based on information that polgen provides.
##     two situations
##  1. patterns - writer is notified that if an environmental resource is treated as special there would be a pattern
##  2. groups of read or modified - default is to bring up the original environment set type.  If policy writer chooses
##    to override the default, we pop up a message - " are you sure you want to do this"
##


        
##################### Fc Generation ###############################

def get_dir_extension_to_use (name, extension):
    dir = os.path.dirname(name)
    file_name = name.split('/')[-1].split('.')
    possible_name = ""
    if extension == 'dir/name':
        use = name
    elif len(file_name) > 1:
        possible_name = file_name[0]
        if extension == 'dir/name.*':
            use = dir + '/' + possible_name + '.*'
        else:
            use = dir + extension[3:]
    else: # no . in string
        if extension == 'dir/name.*':
            use = name + '.*'
        else:
            use = dir + extension[3:]
    return use

def do_fc_generation(name_extension_pairs, temp = False):
    prgm = get_current_program()
    pname = prgm.name
    fileid  = pname + ".fc"
    if temp:
        fileid = 'temp_' + fileid
    fileid_reg = 'fc.html'
    fileid_ref = 'Ref-fc.html'
    do_fc_generation_aux(pname, fileid, name_extension_pairs, text_request)
    do_fc_generation_aux(pname, fileid_reg, name_extension_pairs, html_request)
    #do_fc_generation_aux(pname, fileid_ref, name_extension_pairs, html_request)
    
def do_fc_generation_aux(prgm_name, fileid, name_extension_pairs, printreq):
    fileStr = config.result_file(fileid)
    fc_strm = open(fileStr, "w")
    print "Writing to " + fileStr
    linefeed = printreq.linefeed
    start= printreq.start
    end = printreq.end
    focus = printreq.focus
    end_focus = printreq.end_focus
    intro = printreq.intro
    ending = printreq.ending
    fc_strm.write(intro)
    responses = []
    unresolved = []
    resolved = []
    for item in name_extension_pairs:
        if contains_str(item[0], 'prefix'):
            unresolved.append(item)
        else:
            resolved.append(item)
    for item in unresolved + resolved:

        name = item[0]
        obj = resources_dict[name + ':File']
        context = obj.security_context
        extension = item[1]
        if contains_str(context, '_exec_'):
            extension = 'dir/name'
        for i in range(3):
            if len(context) > i + 1:
                if context[-i -1] == '.':
                    extension = 'dir/name'
        use = get_dir_extension_to_use(name, extension)
        responses.append([use, context])
    responses = seq2set(responses)
    max = 0
    for  r in responses:
        if len(r[0]) > max:
            max = len(r[0])
    delim = ""
    end = ""
    if linefeed != "":
            fc_strm.write('<table> \n')
            linefeed = '<tr><td>'
            delim = '</td><td>'
            end = '</td></tr>'
    for r in responses:
        fname = r[0]
        context = r[1]
        there = len(fname)
        spaces = max - there + 40
        fc_strm.write("%s %s \t\t %s %s gen_context(%s, s0)%s \n" % (linefeed, fname, delim, delim, context, end))
    if linefeed != "":
        fc_strm.write('</table> \n')
    fc_strm.close()

################## Te Generation Strategies ########################
    


def populate_marks_on_transitions():
    for realization in [x for x in pr.all_pattern_realizations.values() if x.active == 1]:
        parts = realization.participants
        walk_participants(realization, parts)

def walk_participants (realization, remainder):
    """ remainder is a list of participants """
    if len(remainder) > 1:
        r1 = remainder[0]
        r2 = remainder[1]
        if r2 in trans[r1].keys():
            trans[r1][r2].marked.append(realization)
        elif r1 in trans[r2].keys():
            trans[r2][r1].marked.append(realization)
        walk_participants(realization, remainder[1:])



def best_choice (nt, calls):
    for cat in module_category_order:
        temp = []
        for c in calls:
            if c.category == cat:
                temp.append(c)
        if len(temp) > 0:
           break
    use = reduce(nt.action, temp)
    return use

def reduce (target, calls, parsimonious = True):
    """target is a string, calls a list of interface calls"""
    # first find calls that only permit a minimum set of permissions
    # each c in calls has an allow with target in list of permissions
    aliases = [target] + actions[target].friends
    lst = []
    for c in calls:
        allows = c.allows
        #find the allow(s)
        #a.perms is a string
        consider_nbr = min([len(a.perms) for a in allows if not \
                            empty([x for x in aliases if x in a.perms])])
        lst.append([c, consider_nbr])
    #now reduce to those calls that have a minimum consider number
    min_perms = 1000
    for c in lst:
        if c[1] < min_perms:
            min_perms = c[1]
    possible = [x[0] for x in lst if x[1] == min_perms]
    toreturn = possible[0]
    min_allows = 1000
    if parsimonious:
        for c in possible:
            nbr = len(c.allows)
            if nbr < min_allows:
                min_allows = nbr
                toreturn = c
    return toreturn

def clean_up ():
    for r in resources_dict.values():
        new_sc = temp_type_replacement(r)
        if new_sc!= r.get_sc():
            TY_Replace().go(r, context2type(new_sc))
        domain = PRGM.unconfined_domain
        if domain != 'unconfined_t':
            #need to change all unconfined_t
            pos = context2type(new_sc).find('unconfined_t')
            if pos != -1:
                TY_Replace().go(r, domain, 'setting unconfined_domain')

################### Fix transitions based on use in patterns or user interactions ####
                
def set_assertion_coverage (final = 0):
    """Actions to adjust coverage of transtions (i.e., covered by pattern or not)
    prior to policy generation"""
    setup_normalized_transitions(final) # re-do with new types
    redo_pattern_assertions()

def redo_pattern_assertions():
    """Populating pattern_assertion_final dictionary"""
    for preal in pr.all_pattern_realizations.values():
        pr.pattern_assertion_final_factory(preal)


#################### start generation ##################
        
def do_policy_generationCB():
    items = [f for f in resources_dict.values() if the_type(f) == 'File']
    data = [[f.name, f.sc_extent] for f in items if \
            f.original_context != f.get_sc() \
            or f.sc_from == 'typegen']
            #f.original_context != f.get_sc()]
    if not get_current_program().using_tracker_data:
        print "Generating temporary .fc, .te, .if files"
        PolgenTemp().do_text_pages()
    do_fc_generation(data, True) 
    clean_up()
    set_assertion_coverage(1)
    problems = multiple_transitions()
    if len(problems) > 0:
        do_anomaly()
        print "\nRepairing multiple transition anomalies"
        for k in problems.keys():
            TY_Repair_Multiple_Trans().go(problems[k])
        print '\n'

    do_fc_generation(data) 
    gen_strategy.do_text_pages()
    gen_strategy.do_web_pages()
    for f in gen_strategy.files.values():
        f.close()


def multiple_transitions ():
    execs = [x for x in pr.all_pattern_realizations.values() \
             if x.pattern.name == 'Executable']
    seen = {}
    for e in execs:
        types = str([context2type(x.get_sc()) for x in e.participants][:-1])
        if types not in seen.keys():
            seen[types] = [e]
        else:
            seen[types].append(e)
    problems = {}
    for k in seen.keys():
        there = seen[k]
        #problem if they don't all have the same third type
        third = seq2set([context2type(x.participants[2].get_sc()) \
                         for x in there])
        if len(third) > 1:
            problems[k + "->" + str(third)] = there
    for k in problems.keys():
        print "Warning: Multiple transition problem at" +   k
    return problems
