import config
from util import *
from domain_model import resource_factory
import sys
from dutil import *
import shutil
import exceptions
import unicodedata

nodes = {}
global LOCAL_INDICATOR
LOCAL_INDICATOR = ""

map2type = {'/dev/tty' : 'devtty_t',
           '/usr/bin/perl' : 'bin_t',
            'lib' : 'lib_t',
            'usr/lib' : 'lib_t',
           "selinux" : 'security_t',
           "/etc/ld.so.cache" : "ld_so_cache_t",
            "/etc/mtab"	: 'etc_runtime_t',
            'CVS/Root': 'cvs_t',
            'prefix' : 'polgen_temp_data_t',
            'var_prefix' : 'polgen_temp_data_t',
            'exec_prefix' : 'polgen_temp_exec_t',
            'USER_SUPPLIED' : 'polgen_temp_t',
            'mta_files' : 'mta_delivery_agent'}

node_properties = ["requires", "activators", "reads", "writes", "execs", "connects_at", \
                    'pipe', 'pipe_reads', 'pipe_writes', 'provides', 'pipeline', \
                   'listens_at']

node_singleton_properties = ['type', 'is_part_of', 'entry_point'] + [qname for qname in questions_dict.keys()]

class PSLError (exceptions.Exception):
    def __init__(self, args = None):
        self.args = args
#-------------------------------------------
# Nodes - the ultimate product produced by the parser

class Node:

    def __init__(self, sobj, name):
        self.created_as_owner = False
        self.pipeline = []
        self.location = 'external'
        self.name = name
        self.source = sobj
        self.activators = [] # set of strings
        self.is_part_of = None
        self.reads= []
        self.entry_point = 0
        self.provides = []
        self.writes = []
        self.listens_at = []

        self.execs = []
        self.type = 'process'
        self.connects_at = []
        for qname in questions_dict.keys():
            setattr(self,qname,False)
        self.requires = []
        nodes[name] = self
    def get_extended_reads (self):
        return self.reads + flatten([nodes[x].get_extended_reads() for x in self.requires]) 
    def get_extended_writes (self):
        return self.writes + flatten([nodes[x].get_extended_writes() for x in self.requires])
    def get_extended_connects_at (self):
        return self.connects_at + flatten([nodes[x].get_extended_connects_at() for x in self.requires])
    def get_extended_execs (self):
        return self.execs + flatten([nodes[x].get_extended_execs() for x in self.requires])
    def get_extended_listens_at (self):
        return self.listens_at + flatten([nodes[x].get_extended_listens_at() for x in self.requires])
    def __repr__ (self):
        return self.name
    def browse(self):
        self.writer(sys.stdout)
    def writer (self, strm):
        strm.write("component " +  self.name + ' {\n')
        for x in node_singleton_properties:
            if self.__dict__[x] != None and self.__dict__[x] != False:
             strm.write("     " + x +  ' ' + str(self.__dict__[x]) + ' \n')            
        for x in node_properties:
            if x in self.__dict__.keys():
                there = self.__dict__[x]
                if Exists(there):
                    strm.write("     " + x +  ' {')
                    nbr = len(there)
                    count = 1
                    for item in there:
                       delim = ', \n   '
                       if count == nbr:
                           delim = ''
                       strm.write(item + delim)
                       count = count + 1
                    strm.write('} \n')
        strm.write('} \n\n')

   
        

#-----------------------------------------------------
# SpecString - intermediate data structure for holding collection
#of text strings

replacement_map = {}

def set_replacement_map(target_pairs, replacement_pairs):
    if len(target_pairs) > 0:
        current = target_pairs[0]
        repl_key = current[0]
        target_str = current[1]
        for rp in replacement_pairs:
                target_str = target_str.replace('$' + rp[0], rp[1])
        if len(target_pairs) > 1:
            replacement_pairs.append([repl_key, target_str])
            set_replacement_map(target_pairs[1:], replacement_pairs)
        replacement_map[repl_key] = target_str
        
class SpecString:
    def __init__(self, specstring, initial = 1):
        self.text = specstring
        self.resources = {}
        self.declarations = {}
        self.components = []
        if initial:
            self.set_component_strings()
        
    def set_component_strings (self):
        replacement_pairs = []
        items = self.text.split('component')
        declarations = items[0].split('\n')
        declaration_pairs = []
        for d in declarations:
            pair = d.split('=')
            if len(pair) == 2:
                declaration_pairs.append([pair[0].strip(), pair[1].strip()])
            elif d != "" and len(pairs) != 0:
                print "d = " + d
                raise PSLError, "Check Declarations in .psl file; missing or multiple '=' on a line"
        global LOCAL_INDICATOR
        LOCAL_INDICATOR = declaration_pairs[0][1]        
        set_replacement_map(declaration_pairs, [])
        raw_components = items[1:]
        for c in raw_components:
            for k in replacement_map.keys():
                c = c.replace('$' + k, replacement_map[k])
            self.components.append(c)
                                         
                                         

#--------------------------------------------------
# Reader Visitors

class ReaderVisitor:
    def __init__(self, reader):
        self.target = reader

class SpecParser (ReaderVisitor):
    def do_it (self, sobj):
        strm = self.target.out
        components = self.target.specobj.components
        for c in components:
            start = c.strip()
            name = start.split()[0]
            utype = unicodedata.category(unicode(name[0]))
            if utype == 'Nd':
                name = 'Service_at_' + name
            #catch multiple entries
            if name in nodes.keys() and not nodes[name].created_as_owner\
               and not name == get_current_program().name and \
               nodes[name].location == 'internal':
                raise PSLError, "Multiple components with the same name = " + name
                return False
            co = node_factory(sobj, name, 'process')
            global LOCAL_INDICATOR
            if name.find(LOCAL_INDICATOR) == 0:
                co.location = 'internal'
            attributes = start[len(name) + 2:-1].strip()
            begin = True
            vals = []
            next = None
            for a in attributes.split():
                if begin:
                    aname = a
                    if next in ['is_part_of', 'type', 'entry_point'] + questions_dict.keys():
                        setattr(co, next, aname)
                        next = None                            
                    elif aname == '{':
                        continue
                    elif aname in ['is_part_of', 'type', 'entry_point'] + questions_dict.keys():
                        next = aname
                    elif aname not in node_properties + node_singleton_properties:
                        print attributes
                        raise PSLError, "Undefined attribute " + aname
                        return False
                    else:
                        begin = False               
                else:
                    if a == ' ': # should not happen
                        continue
                    if a[-1] == '\n':
                        a = a[:-1]
                    items = a.split(',')
                    if items[-1] == '':
                        items = items[:-1]
                    if len(a) == 0:
                        continue
                    elif a[0] == '{': # strip off the { from the first item
                        items = [items[0][1:]] + items[1:]
                        if len(a) == 1:
                            continue
                        else:
                           for x in items:
                               if x[-1] == '}':
                                   x = x[:-1]
                                   if len(x) > 0:
                                    vals.append(fix_value(aname, x))
                           if a[-1] == '}':
                                begin = True
                                sput(co, aname, vals)
                                vals = []
                    elif a[-1] == '}':
                                begin = True
                                for x in items:
                                    if x[-1] == '}':
                                        x = x[:-1]
                                    if len(x) > 0:
                                        vals.append(fix_value(aname, x))
                                sput(co, aname, vals)
                                vals = []
                             
                    else:
                        for x in items:
                                vals.append(fix_value(aname, x))
        for n in nodes.values():
            if n.is_part_of != None:
                if type(n.is_part_of) == type(""):
                    if n.is_part_of in nodes.keys():
                        parent = nodes[n.is_part_of]
                    else: #create it now; remove order dependency
                        parent = Node(sobj, n.is_part_of)
                        parent.created_as_owner = True
                    n.is_part_of = parent

class Spec2Html (ReaderVisitor):
    def do_it (self, sobj):
        strm = self.target.out
        strm.write('<h1>' + get_current_program().name + ' Specification </h1> \n')
        components = self.target.specobj.components
        strm.write('<html><table border = 1 width = 80%>\n')
        for c in components:
            strm.write('<tr><td>')
            start = c.strip()
            name = start.split()[0]
            strm.write('Component <a href = PolicyHelp.html#' + name + '>' + name + '</a> \n') 
            count = 0
            attributes = start[len(name) + 2:-1].strip()
            for a in attributes.split():
                start = '<br>----'
                if count == 1:
                    count = 0
                    start = " "
                else:
                    count = 1
                strm.write(start + a + ' \n')
            strm.write('</td></tr>')
        strm.write('</table> \n\n')
        strm.close()

class PslWriter:
    def __init__(self):
        """ Copies psl file and writes new one from current nodes 
         """
        prgm = get_current_program().name
        base = config.data_directory + '/'
        pathname = base + prgm + '.psl'
        s = "psl"
        files = [x for x in os.listdir(base) if x[-(len(s) + 1):] == "." + s]
        for f in files:
            out_file = base + f + '_bkp'
            print "Writing to " + out_file
            try:
                shutil.move(base + f, out_file)
            except:
                print "Problem copying a file"
        # create one large psl file for subsequent runs
        self.file = open(pathname, 'w')
        self.file.write("# declarations \n")
        for n in nodes.values():
            if n.type != 'data_file' and n.location == 'internal':
                n.writer(self.file)
        self.file.close()
        

        

#------------------------------------------------
# Readers

class Reader:
    def __init__(self, filename, folder = ""):
        """ Initializes the class by opening the file
         """
        if folder == "":
            add = '/'
        else:
            add = "/" + folder + "/"
        pathname = config.data_directory + add + filename
        out_file = config.results_directory + "/" \
        + filename + '.html'
        print "Reading from data in " + pathname
        self.file = open(pathname, 'r')
        self.out= open(out_file, 'w')
        self.first_line = 1
        self.specobj = None
        self.name = filename
        
class PSLReader (Reader):
    """A class for reading psl files."""
    def readlines (self, sobj):
        file = self.file
        spec = ""
        for l in file.readlines():
            if len(l.strip()) > 0 and l.strip()[0] == "#":
                continue
            spec = spec + l
        self.specobj = SpecString(spec)
        file.close()



def fix_value(aname, val):
    return_val = val
    if aname[-5:] == 'types':
        return_val = []
        if val[-2:] == '_t':
            return_val = val
        else:
            return_val = val + '_t'
    return return_val

def sput (obj, slot, val):
    setattr(obj, slot,val)
    if slot in ['reads', 'writes']:
        #create a data node
        for item in val:
            n = node_factory(obj.source, item, 'data_file')
    if slot in ['execs', 'connects_at', 'listens_at']:
        for item in val:
            target = node_factory(obj.source, item, 'process')
            #capturing the back pointers 
            if slot == 'execs':
                target.activators.append(obj.name)
    elif slot == 'requires':
        for item in val:
            node_factory(obj.source, item, 'code_module')

def node_name (name):
    utype = unicodedata.category(unicode(name[0]))
    if utype == 'Nd':
        name = "Service_at_" + name
    return name

def node_factory(sobj, name, type):
    toreturn = None
    name = node_name(name)
    if name in nodes.keys():
        obj = nodes[name]
        otype = obj.type
        if otype == type:
            toreturn = obj
        elif type == 'code_module' and otype in ['process', 'entry_point', 'daemon', 'application']:
            toreturn = obj
        elif otype in ['application', 'entry_point', 'daemon'] and type == 'process':
            toreturn = obj
        else:
            print "Name clash with " + name + ": " + obj.type + " versus type " + type
    else:
        toreturn = Node(sobj, name)
        toreturn.type = type
        global LOCAL_INDICATOR
        if name.find(LOCAL_INDICATOR) == 0:
            toreturn.location = 'internal'
    return toreturn



#preliminary - will need work
def psl2xml (file):
    out_pathname = config.data_directory + '/' + \
                   file[:-4] + '.xml'
    strm = open(out_pathname, 'w')
    text = read_it(file)
    parse_it(text) # creates nodes dictionary
    strm.write('<components> \n')
    for item in nodes.values():
        strm.write('<component> \n')
        for k in item.__dict__.keys():
            there = item.__dict__[k]
            nbr = 1
            mults = False
            if type(there) == type([]):
                nbr = len(there)
                mults = True
            if nbr > 0:
                strm.write("   <" + k + "> \n")
                count = 0
                if mults:                                            
                    for v in item.__dict__[k]:
                        count = count + 1
                        add = ','
                        if count == nbr:
                            add = ''
                        strm.write('    ' + str(v) + add + '\n')
                else:
                    strm.write('     ' + str(item.__dict__[k]) + '\n')
                strm.write("   </" + k + "> \n")
        strm.write('</component> \n')
    strm.write('</components>')
    strm.close()

## obsolete - not called
def post_process ():
    for c in nodes.values():
        if c.is_part_of != None:
            c.is_part_of = nodes[c.is_part_of]
        for x in c.socket_write:
            c.socket_write_obj.append(nodes[x])
        for x in c.socket_read:
            c.socket_read_obj.append(nodes[x])

def get_resource_name (givenstr):
    #for now
    return givenstr



def file_sc_from_type (givenstr):
    start = 'system_u:object_r:'
    if givenstr[-2:] == '_t':
        completion = givenstr
    elif contains_str(givenstr, get_current_program().name_for_type)\
         or givenstr.find('/tmp') == 0 \
         or nodes[givenstr].location == 'internal':
        completion = 'polgen_temp_data_t'
    elif givenstr not in map2type.keys() and nodes[givenstr].location == 'external':
        dtrace(givenstr)
        completion = givenstr + '_t'
    else:
        found = False
        #do special cases first
        for k in map2type.keys():
            if givenstr.find(k) == 0:
                completion = map2type[k]
                found = True
                break
        if not found:
            #do defaults for head directories
            if len(givenstr) < 5:
                completion = givenstr + '_t'
            elif contains_str(givenstr, '/lib'):
                completion = 'shlib_t'
            elif givenstr.find('/etc/') == 0:
                ###cancatenation of subdirectories
                dirs = givenstr.split('/')
                if len(dirs) < 4:
                    completion = 'etc_t'
                else:
                    completion = ""
                    for i in range(len(dirs) - 3):
                        completion = completion + dirs[i + 2] + '_'
                    completion = completion[:-1] + '_t'
            elif givenstr[:5] == '/dev/':
                completion = givenstr[5:] + '_t'
            else:
                #well, I don't know
                completion = 'polgen_temp_data_t'
    return start + completion

def get_process_sc_from_type(givenstr):
    """note: sc not cached on node objects, so this will be called several times"""
    base = 'system_u:object_r:'
    ty = 'bin_t'
    top = get_current_program().name
    if givenstr == top:
        ty = 'polgen_temp_exec_t'
    elif nodes[givenstr].location == 'internal':
        ty = 'polgen_temp_t'
        if nodes[givenstr].entry_point == "1":
            local = get_local_name(givenstr)
            ty = 'polgen_temp_' + local + '_t'
    sc = base + ty
    return sc

global akop_seen

def akop (n1, n2):
    global akop_seen
    """ true if n1 is a subcomponent of n2"""
    akop_seen = [n1]
    return akop_aux(n1, n2)

def akop_aux(n1, n2):
    global akop_seen
    start = nodes[n1]
    if start.name == n2:
        return True
    elif start.is_part_of != None:
        next = start.is_part_of.name
        if next in akop_seen:
            raise PSLError, "Loop in spawning hierarchy"
            return False
        else:
            akop_seen.append(next)
        return akop_aux(next, n2)
    else:
        return False
        
def setup_process (nodeobj, pid, startobj = None):
    """sobj = source, node is a component, pid is the
    process id"""
    pname = get_resource_name(nodeobj.name)
    sobj = nodeobj.source
    base_sc = get_process_sc_from_type(pname)
    process = resource_factory(sobj, pname, 'Process', base_sc)
    if nodeobj.pipeline != []:
        process.internal_pipeline = True
    process.entry_point = nodeobj.entry_point
    process.pid = pid
    # need to create the exec file for the process
    #adding subprocesses to execs of an activator
    #this will set things up so that an exec_t file will be created
    if startobj == None:
        #we're at the top
        if pname not in nodeobj.execs:
            nodeobj.execs.append(pname) # starter of process
    elif nodeobj.activators != []:
        for a in nodeobj.activators:
            aobj = node_factory(sobj, a, 'process')
            if pname not in aobj.execs:
                aobj.execs.append(pname) 
    if nodeobj.type == 'daemon':
        process.special_calls.append('setsid')
    process.processes[pid] = process
    return process

def get_local_name(pname):
    sans_ext = pname.split('.')[0]
    return sans_ext.split('/')[-1]
                
def do_execs (obj, startobj):
    """pobj is a process or file, startobj is the top process"""
    pobj = None
    if startobj != None:
        pname = get_resource_name(obj.name)
        sobj = nodes[pname].source
        if the_type(obj) == 'Process':
            pobj = obj
            startobj.processes[pobj.pid] = pobj
            filename = pname
            new_sc = file_sc_from_type(startobj.name)[:-2] + '_exec_t'
            node = nodes[pname]
            if node.entry_point == "1":
                local_name = get_local_name(pname)                
                new_sc = startobj.security_context[:-2] + '_' + local_name + '_exec_t'
            fileobj = resource_factory(sobj, filename, 'File', new_sc)
            when_exec(sobj, startobj, pobj, filename)
        elif the_type(obj) == 'File':
            filename = obj.name
            pobj =  resource_factory(sobj, filename, 'Process', \
                                     get_process_sc_from_type(pname))
            if the_type(pobj) != 'Process':
                print 'Error: '+ filename + ' not a process'
            if the_type(startobj) != 'Process':
                print 'Error: Startobj '+ startobj.name + ' not a process'
            if pobj.pid == 0:
                pobj.pid = new_pid()
            startobj.processes[pobj.pid] = pobj
            when_exec(sobj, startobj, pobj, filename)
        else:
            print "Exec'ed object is not a process or file"
    return pobj

def new_pid ():
    the_prgm = get_current_program()
    there = the_prgm.next_pid
    the_prgm.next_pid = there + 1
    return there

def when_exec (sobj, responsible_process, process, filestr):
    if the_type(responsible_process) != 'Process':
        print '**When exec - source ' + str(responsible_process) + ' not process'
    if the_type(process) != 'Process':
        print '**When exec - target ' + str(process) + ' not process'
    file = resources_dict[filestr + ':File']
    responsible_process.add_transition('read', filestr, 'File')
    responsible_process.add_transition('exec', filestr, 'File')
    responsible_process.executing.append(process)
    process.executed_by.append(responsible_process)
    #responsible_process.add_transition('exec', filestr, 'File')
    link_name = responsible_process.name + '-directly-execs-' + process.name
    spawn_link = resource_factory(sobj, link_name, 'Spawn', responsible_process.get_sc())
    responsible_process.add_transition('activate', link_name, 'Spawn')
    process.add_transition('spawn', link_name, 'Spawn')
    
                
